from unittest import TestCase
from unittest.mock import MagicMock

from ikcms.forms import exceptions
from ikcms.forms import fields


class BaseTestCase(TestCase):

    field_cls = fields.Base
    type_error_values = []
    test_values = [
        ('test_str', 'test_str'),
        (5, 5),
        ({'aa', 22}, {'aa', 22}),
        ([3, 6], [3, 6]),
    ]
    to_python_default_values = ['test_str', None, 5, [3, 6]]


    validation_tests = []


    def test_init(self):
        context = MagicMock()
        parent = MagicMock()
        field = self.field()(context, parent)
        self.assertEqual(field.context, context.copy())
        self.assertEqual(field.parent, parent)

    def test_to_python(self):
        field = self.field()()
        for python_value, raw_value in self.test_values:
            self.assertEqual(
                field.to_python(raw_value),
                python_value,
                self._to_python_message(raw_value, python_value),
            )
        for value in self.type_error_values:
            with self.assertRaises(exceptions.RawValueTypeError) as ctx:
                field.to_python(value)
            self.assertEqual(ctx.exception.kwargs['field_name'], field.name)

        for kwargs, allowed_values, denied_values in self.validation_tests:
            field = self.field(**dict(kwargs, ))()
            for value in allowed_values:
                field.to_python(value)
            for value in denied_values:
                with self.assertRaises(exceptions.ValidationError):
                    field.to_python(value)

    def test_to_python_default(self):
        field = self.field(raw_required=False)()
        self.assertEqual(field.to_python(fields.NOTSET), fields.NOTSET)
        for value in self.to_python_default_values:
            field = self.field(to_python_default=value, raw_required=False)()
            self.assertEqual(field.to_python(fields.NOTSET), value)

    def test_raw_required(self):
        field = self.field()()
        with self.assertRaises(exceptions.RawValueRequiredError) as ctx:
            field.to_python(fields.NOTSET)
        self.assertEqual(ctx.exception.kwargs['field_name'], field.name)

    def test_from_python(self):
        field = self.field()()
        for python_value, raw_value in self.test_values:
            self.assertEqual(field.from_python(python_value), raw_value)

    def field(self, **kwargs):
        kwargs.setdefault('name', 'test')
        return type('TestField', (self.field_cls,), kwargs)

    def _to_python_message(self, raw_value, value):
        return "{}().to_python({})!={}".format(
            self.field_cls.__name__,
            raw_value,
            value,
        )

    def _from_python_message(self, value, raw_value):
        return "{}().from_python({})!={}".format(
            self.field_cls.__name__,
            value,
            raw_value,
        )


class FieldTestCase(BaseTestCase):

    field_cls = fields.Field

    def test_to_python(self):
        field = self.field()()
        for python_value, raw_value in self.test_values:
            self.assertEqual(
                field.to_python(self.test_dict(field.name, raw_value)),
                {field.name: python_value},
            )

        for value in self.type_error_values:
            with self.assertRaises(exceptions.RawValueTypeError) as ctx:
                field.to_python(self.test_dict(field.name, value))
            self.assertEqual(ctx.exception.kwargs['field_name'], field.name)

        for kwargs, allowed_values, denied_values in self.validation_tests:
            field = self.field(**dict(kwargs, ))()
            for value in allowed_values:
                field.to_python(self.test_dict(field.name, value))
            for value in denied_values:
                with self.assertRaises(exceptions.ValidationError) as ctx:
                    field.to_python(self.test_dict(field.name, value))
                self.assertEqual(
                    list(ctx.exception.kwargs['error'].keys()),
                    [field.name],
                )


    def test_to_python_default(self):
        field = self.field(raw_required=False)()
        self.assertEqual(field.to_python(self.test_dict()), {})

        for value in self.to_python_default_values:
            field = self.field(to_python_default=value, raw_required=False)()
            self.assertEqual(
                field.to_python(self.test_dict()),
                {field.name: value},
            )

    def test_raw_required(self):
        field = self.field(raw_required=True)()
        with self.assertRaises(exceptions.RawValueRequiredError) as ctx:
            field.to_python(self.test_dict())
        self.assertEqual(ctx.exception.kwargs['field_name'], field.name)

    def test_from_python(self):
        field = self.field()()
        for python_value, raw_value in self.test_values:
            self.assertEqual(
                field.from_python(self.test_dict(field.name, python_value)),
                {field.name: raw_value},
            )

    def test_dict(self, name=None, value=None):
        d = {
            'other1': 'str value',
            'other2': 7777,
            'other3': None,
            5: 'this not wrong',
            None: 'xxx',
        }
        if name:
            d[name] = value
        return d


class StringTestCase(FieldTestCase):

    field_cls = fields.String

    test_values = {
        ('test_str', 'test_str'),
        ('', ''),
    }
    type_error_values = [
        5, {'d': 'teet'}, [2, 4], 66.2, set((4, 5, 6)), (3, 4), MagicMock(),
    ]
    validation_tests = [
        ({'required': True}, ['test_str', '325566'], ['']),
        ({'min_len': 3}, ['test_str', '325', '333'*40], ['', '1', '23']),
        ({'max_len': 3}, ['', '3', '32', '123'], ['1234', '23'*5]),
        (
            {'regex': '[abs]{3}$'},
            ['bsa', 'abs'],
            ['ab', 'abd', 'absabs', ''],
        ),
    ]


class IntTestCase(FieldTestCase):

    field_cls = fields.Int

    test_values = [
        (1, 1),
        (-4, -4),
        (500, 500),
        (0, 0),
    ]
    type_error_values = [
        '5', {'d': 'teet'}, [2, 4], 66.2, set((4, 5, 6)), (3, 4), MagicMock(),
        'dddddd',
    ]
    validation_tests = [
        ({'required': True}, [5, 10, -6, 0], []),
        ({'min_value': 3}, [3, 325, 333*40], [-50, 0, 1, 2]),
        ({'max_value': 3}, [-50, 0, 1, 2, 3], [4, 10, 104, 5*600]),
    ]


class IntStrTestCase(FieldTestCase):

    field_cls = fields.IntStr

    test_values = [
        (1, '1'),
        (-4, '-4'),
        (500, '500'),
        (0, '0'),
    ]
    type_error_values = [
        5, 0, -6, 66.2,
        {'d': 'teet'}, [2, 4], set((4, 5, 6)), (3, 4), MagicMock(),
    ]
    validation_tests = [
        ({'required': True}, ['5', '10', '-6', '0'], ['asss', '5.5', '66sx']),
        ({'min_value': 3}, ['3', '325'], ['-50', '0', '1', '2']),
        ({'max_value': 3}, ['-50', '0', '1', '2', '3'], ['4', '10', '104']),
    ]



class DictTestCase(FieldTestCase):

    field_cls = fields.Dict
    type_error_values = [5, [2, 4], "ddddd", MagicMock()]
    test_values = [
        (
            {'f1': 'value1', 'f2': 'value2'},
            {'f1': 'value1', 'f2': 'value2'},
        ),
        (
            {'f1': 'value1', 'f2': None},
            {'f1': 'value1', 'f2': None},
        ),

    ]
    validation_tests = [
        (
            {'required': True},
            [
                {
                    'f1': 'value1',
                    'f2': 'value2',
                    'other4': 'value3',
                },
                {
                    'f1': 'value1',
                    'f2': 'value2',
                },
            ],
            []
        ),
    ]

    def field(self, **kwargs):
        class f1(fields.Field):
            name = 'f1'
        class f2(fields.Field):
            name = 'f2'
            not_none = False
        kwargs.setdefault('fields', [f1, f2])
        return super().field(**kwargs)


class ListTestCase(FieldTestCase):

    field_cls = fields.List
    type_error_values = [5, {'x':'xxx'}, "ddddd", MagicMock()]
    test_values = [
        ([], []),
        ([1, 2, 3], [1, 2, 3]),
        ([1, 3], [1, 3]),
        ([{}, [], []], [{}, [], []]),
    ]

    validation_tests = [
        (
            {'required': True},
            [[1, 2, 3], [[]]],
            [[]],
        ),
    ]

    def field(self, **kwargs):
        kwargs.setdefault('fields', [fields.Field])
        return super().field(**kwargs)


class BlockTestCase(BaseTestCase):

    field_cls = fields.Block
    type_error_values = [
        {'test': 5},
        {'test': [2, 4]},
        {'test': "ddddd"},
        {'test': MagicMock()},
    ]
    to_python_default_values = []

    test_values = [
        (
            {
                'f1': 'value1',
                'f2': 'value2'
            },
            {
                'test': {
                    'f1': 'value1',
                    'f2': 'value2',
                },
            },
        ),
        (
            {
                'f1': 'value1',
                'f2': None,
            },
            {
                'test': {
                    'f1': 'value1',
                    'f2': None,
                }
            }
        ),
    ]
    validation_tests = [
        (
            {
                'required': True,
            },
            [
                {
                    'test': {
                        'f1': 'value1',
                        'f2': 'value2',
                        'other4': 'value3',
                    },
                },
                {
                    'test': {
                        'f1': 'value1',
                        'f2': 'value2',
                    },
                },
            ],
            [],
        ),
    ]

    def field(self, **kwargs):
        class f1(fields.Field):
            name = 'f1'
        class f2(fields.Field):
            name = 'f2'
            not_none = False
        kwargs.setdefault('fields', [f1, f2])
        return super().field(**kwargs)

    def test_to_python_default(self):
        field = self.field(raw_required=False)()
        self.assertEqual(field.to_python({}), {})

        for value in self.to_python_default_values:
            field = self.field(to_python_default=value)()
            self.assertEqual(field.to_python({}), value)

    def test_raw_required(self):
        field = self.field(raw_required=True)()
        with self.assertRaises(exceptions.RawValueRequiredError) as ctx:
            field.to_python({})
        self.assertEqual(ctx.exception.kwargs['field_name'], field.name)





import logging
import cPickle as pickle
import time

from sqlalchemy.orm import Query
from sqlalchemy import func
from sqlalchemy import sql
import sqlalchemy.exc

import ikcms.components.base
from ikcms.web import h_cases
from ikcms.utils import cached_property

from . import views

logger = logging.getLogger(__name__)


class Component(ikcms.components.base.Component):

    name = 'sections'
    model = 'front.Section'
    check_timeout = 3
    lock_timeout = 3
    handler_updated_ts = None
    views = {
        'dir': views.DirView,
        'page': views.PageView,
        'view': views.PageView,
    }
    custom_views = {}


    def __init__(self, app):
        super(Component, self).__init__(app)
        self.model = self.app.db.get_model(self.model)
        self.cache_key_checked_ts = '{}:checked_ts'.format(self.name)
        self.cache_key_updated_ts = '{}:updated_ts'.format(self.name)
        self.cache_key_updating = '{}:updating'.format(self.name)
        self.cache_key_meta = '{}:meta'.format(self.name)
        self.cache_key_body = '{}:body'.format(self.name)
        self.cache_key_lock = '{}:lock'.format(self.name)
        self.init_cache()

    def on_request(self, request):
        # Update handler and root, if sections was updated
        cache_updated_ts = self.update_cache()
        if cache_updated_ts != self.handler_updated_ts:
            self.handler_updated_ts = cache_updated_ts
            self.app.handler = self.app.get_handler()
            self.app.root = self.app.get_root()

    def reset_cache(self, pipe):
        self.app.cache.delete(
            self.cache_key_updated_ts,
            self.cache_key_updating,
            self.cache_key_meta,
            self.cache_key_body,
        )

    def init_cache(self):
        with self.app.cache.lock(
            self.cache_key_lock,
            expires=self.lock_timeout,
        ) as lock:
            self.update_cache()

    def update_cache(self):
        # check update timeout
        cache_checked_ts, cache_updated_ts = self.app.cache.mget(
            self.cache_key_checked_ts,
            self.cache_key_updated_ts,
        )
        try:
            cache_checked_ts = int(cache_checked_ts)
            cache_updated_ts = int(cache_updated_ts)
        except (TypeError, ValueError):
            cache_checked_ts = None
            cache_updated_ts = None

        # if check timeout not expired, updating is not required
        if cache_checked_ts:
            if time.time() < (cache_checked_ts + self.check_timeout):
                return cache_updated_ts

        # if other process is already updating cache, we do nothing
        if not self.app.cache.add(self.cache_key_updating, 1, self.lock_timeout):
            return cache_updated_ts

        now_ts = int(time.time())
        try:
            db_updated_ts = self.get_updated_ts_from_db()
        except sqlalchemy.exc.DBAPIError as exc:
            logger.warning('Retrieve sections error: {}'.format(exc))
            return None

        # if db not changed, we update cache checked ts
        if cache_updated_ts and db_updated_ts <= cache_updated_ts:
            with self.app.cache.pipe() as pipe:
                pipe.set(self.cache_key_checked_ts, now_ts)
                pipe.delete(self.cache_key_updating)
                pipe.execute()
            return cache_updated_ts

        # Get sections from db
        try:
            sections_meta, sections_body = self.get_sections_from_db()
        except sqlalchemy.exc.DBAPIError as exc:
            logger.warning('Retrieve sections error: {}'.format(exc))
            return None

        sections_meta = {s_id: self._dumps(s) \
            for s_id, s in sections_meta.items()}
        sections_body = {s_id: self._dumps(s) \
            for s_id, s in sections_body.items()}

        # Update cache 
        with self.app.cache.pipe() as pipe:
            pipe.set(self.cache_key_updated_ts, db_updated_ts)
            pipe.set(self.cache_key_checked_ts, now_ts)
            pipe.delete(self.cache_key_updating)
            pipe.delete(self.cache_key_meta)
            pipe.delete(self.cache_key_body)
            pipe.hmset(self.cache_key_meta, sections_meta)
            if sections_body:
                pipe.hmset(self.cache_key_body, sections_body)
            pipe.execute()
        return db_updated_ts

    def get_updated_ts_from_db(self):
        session = self.app.db()
        s = sql.select([func.max(self.model.updated_dt)])
        result = list(session.execute(s))
        session.close()
        if result and result[0][0]:
            return int(time.mktime(result[0][0].timetuple()))
        else:
            return None

    def get_sections_from_db(self):
        session = self.app.db()
        sections_objs = session.query(self.model).\
            order_by(self.model.order.asc()).all()

        objs_by_id = {obj.id: obj for obj in sections_objs}

        sections = [section.to_meta_dict() for section in sections_objs \
            if section.slug]
        sections_by_id = {section['id']: section for section in sections}
        sections_by_parent = {}
        for section in sections:
            if section['parent_id'] is not None:
                parent_section = sections_by_id.get(section['parent_id'])
                if not parent_section:
                    continue
            sections = sections_by_parent.setdefault(section['parent_id'], [])
            if section['slug'] not in [s['slug'] for s in sections]:
                sections.append(section)

        def walk_sections(sections_by_parent, parent_section=None):
            result = []
            if parent_section is None:
                parent_id = None
                parent_path = []
                parents = []
            else:
                parent_id = parent_section['id']
                parents = parent_section['parents'] + [parent_section['id']]
                parent_path = parent_section['path']
            for section in sections_by_parent.get(parent_id, []):
                used_slugs = set()
                if section['slug'] in used_slugs:
                    continue
                else:
                    used_slugs.add(section['slug'])
                section = section.copy()
                section['parents'] = list(parents)
                section['path'] = parent_path + [section['slug']]
                children = sections_by_parent.get(section['id'], [])
                section['children'] = [c['id'] for c in children]
                result.append(section)
                result += walk_sections(sections_by_parent, section)
            return result

        sections = walk_sections(sections_by_parent)
        sections_meta = {section['id']: section for section in sections}
        sections_body = {sid: objs_by_id[sid].to_body_dict()\
            for sid in sections_meta}
        root_sections_ids = [s['id'] for s in sections_by_parent.get(None, [])]
        sections_meta[''] = {'children': root_sections_ids}
        session.close()
        return sections_meta, sections_body

    def set_section_meta(section, section_obj):
        return section

    def get_sections(self, ids):
        return self._get_sections_meta(ids)

    def get_section(self, id):
        return self.get_sections([id])[0]

    def get_sections_with_body(self, db, ids):
        with self.app.cache.pipe() as pipe:
            metas = self._get_sections_meta(ids)
            bodies = self._get_sections_bodies(ids)
        sections = []
        for meta, body in zip(metas, bodies):
            if meta is not None and body is not None:
                section = dict(meta, **body)
            else:
                section = None
            sections.append(section)
        return sections

    def get_section_with_body(self, db, id):
        return self.get_sections_with_body(db, [id])[0]

    def get_subsection_by_slugs(self, slugs, section=None):
        # cache?
        section = section or self.get_section('')
        for slug in slugs:
            subsections = self.get_sections(section['children'])
            for subsection in subsections:
                if subsection['slug'] == slug:
                    section = subsection
                    break
            else:
                return None
        return section


    def h_subsections(self, section):
        handlers = []
        if section is None:
            subsections = []
        else:
            subsections = self.get_sections(section['children'])
        for subsection in subsections:
            if not subsection:
                continue

            path = '.'.join(subsection['path'])
            view  = self.custom_views.get(path)
            if not view:
                view = self.views[subsection['type']]
            handler = view.handler(self, subsection)
            handlers.append(handler)
        return h_cases(*handlers)

    def h_sections(self):
        section = self.get_section('')
        return self.h_subsections(section)

    def _get_sections_meta(self, ids):
        if not ids:
            return []
        raw_sections = self.app.cache.hmget(self.cache_key_meta, ids)
        sections = []
        for id, section in zip(ids, raw_sections):
            if section is not None:
                section = self._loads(section)
            sections.append(section)
        return sections

    def _get_sections_bodies(self, ids):
        if not ids:
            return []
        raw_sections = self.app.cache.hmget(self.cache_key_body, ids)
        sections = []
        for id, section in zip(ids, raw_sections):
            if section is not None:
                section = self._loads(section)
            sections.append(section)
        return sections

    def _dumps(self, obj):
        return pickle.dumps(obj)

    def _loads(self, string):
        return pickle.loads(string)


component = Component.create_cls



import ikcms.components.render.jinja2.custom_tags


class IktomiMediaTag(ikcms.components.render.jinja2.custom_tags.Tag):

    item_field_name = 'medias'
    tags = ['iktomi_media']
    template = 'tags/iktomi_{media.type}'
    media_types = ('photo', 'photoset', 'video')

    def func(self, env, item, **kwargs):
        try:
            media_id = int(kwargs.get('item_id'))
        except (TypeError, ValueError):
            return None
        collection = getattr(item, self.item_field_name)
        media = next((b for b in collection if b.id == media_id), None)
        if media:
            assert media.type in self.media_types, \
                'Unknown media type {}'.format(media.type)
            return dict(
                env=env,
                item=item,
                media=media,
                **kwargs
            )
        else:
            return None


class IktomiPhotoTag(IktomiMediaTag):

    item_field_name = 'photos'
    tags = ['iktomi_photo']
    template = 'tags/iktomi_photo.html'
    media_types = ('photo',)


class IktomiPhotosetTag(IktomiMediaTag):

    item_field_name = 'photosets'
    tags = ['iktomi_photoset']
    template = 'tags/iktomi_photoset.html'
    media_types = ('photoset',)


class IktomiVideoTag(IktomiMediaTag):

    item_field_name = 'videos'
    tags = ['iktomi_video']
    template = 'tags/iktomi_video.html'
    media_types = ('video',)


iktomi_media = IktomiMediaTag
iktomi_photo = IktomiPhotoTag
iktomi_photoset = IktomiPhotosetTag
iktomi_video = IktomiVideoTag



from unittest import TestCase, skipIf

from sqlalchemy import sql
from ikcms.utils.asynctests import asynctest
from ikcms.ws_components.db import Component
from ikcms.orm.session import Session
from ikcms.orm import exc

from tests.cfg import cfg
from tests.models import create_models1, create_models2

try:
    import aiomysql
    mysql_skip = False
except ImportError:
    mysql_skip = True

try:
    raise ImportError
    import aiopg
    pg_skip = False
except ImportError:
    pg_skip = True


class _SessionTestCaseBase(TestCase):

    test_models = {
        'db1': create_models1(),
        'db2': create_models2(),
    }

    async def asetup(self):
        engines = {
            'db1': await Component.create_engine('db1', self.db_url),
            'db2': await Component.create_engine('db2', self.db_url),
        }
        test_table1 = self.test_models['db1'].test_table1
        test_table2 = self.test_models['db2'].test_table2
        binds = {
            test_table1: engines['db1'],
            test_table2: engines['db2'],
        }

        async with engines['db1'].acquire() as conn:
            await self.test_models['db1'].reset(conn)

        async with engines['db2'].acquire() as conn:
            await self.test_models['db2'].reset(conn)

        return {
            'engines': engines,
            'binds': binds,
        }

    async def aclose(self, engines, binds):
        for engine in engines.values():
            engine.terminate()
            await engine.wait_closed()

    @asynctest
    async def test_get_engine(self, engines, binds):
        test_table1 = self.test_models['db1'].test_table1
        test_table2 = self.test_models['db2'].test_table2
        async with Session(engines, binds) as session:
            i1 = sql.insert(test_table1)
            i2 = sql.insert(test_table2)
            u1 = sql.update(test_table1)
            u2 = sql.update(test_table2)
            d1 = sql.delete(test_table1)
            d2 = sql.delete(test_table2)
            s1 = sql.insert(test_table1)
            s2 = sql.insert(test_table2)
            for q1, q2 in [(i1, i2), (u1, u2), (d1, d2), (s1, s2)]:
                engine1 = session.get_engine(q1)
                engine2 = session.get_engine(q2)
                self.assertEqual(engine1, engines['db1'])
                self.assertEqual(engine2, engines['db2'])
            with self.assertRaises(exc.OrmError):
                session.get_engine('error query')

    async def _test_execute(self, engines, binds):
        test_table1 = self.test_models['db1'].test_table1
        test_table2 = self.test_models['db2'].test_table2
        async with Session(engines, binds) as session:
            q = sql.insert(test_table1).values(id=5, title='test_title')
            result = await session.execute(q)
            self.assertEqual(result.lastrowid, 5)
            q = sql.select(test_table1.c).where(test_table1.c.id == 5)
            result = await session.execute(q)
            self.assertEqual(result.rowcount, 1)
            result = list(result)
            self.assertEqual(result[0]['id'], 5)
            self.assertEqual(result[0]['title'], 'test_title')

            q = sql.update(test_table1).where(test_table1.c.id == 5).\
                    values(title='test_title2')
            result = await session.execute(q)
            self.assertEqual(result.rowcount, 1)
            q = sql.select(test_table1.c).\
                    where(test_table1.c.id == 5)
            result = await session.execute(q)
            self.assertEqual(result.rowcount, 1)
            result = list(result)
            self.assertEqual(result[0]['id'], 5)
            self.assertEqual(result[0]['title'], 'test_title2')

            q = sql.delete(test_table1).where(test_table1.c.id == 5)
            result = await session.execute(q)
            self.assertEqual(result.rowcount, 1)
            q = sql.select(test_table1.c).\
                    where(test_table1.c.id == 5)
            result = await session.execute(q)
            self.assertEqual(result.rowcount, 0)

    @asynctest
    async def test_rollback(self, engines, binds):
        test_table1 = self.test_models['db1'].test_table1
        test_table2 = self.test_models['db2'].test_table2
        async with Session(engines, binds) as session:
            q = sql.insert(test_table1).values(id=5, title='test_title')
            await session.execute(q)
            q = sql.insert(test_table2).values(id=10, title='test_title2')
            await session.execute(q)
            await session.rollback()
        async with Session(engines, binds) as session:
            q = sql.select(test_table1.c).where(test_table1.c.id == 5)
            rows = await session.execute(q)
            self.assertEqual(rows.rowcount, 0)
            q = sql.select(test_table2.c).where(test_table2.c.id == 10)
            rows = await session.execute(q)
            self.assertEqual(rows.rowcount, 0)

        try:
            async with Session(engines, binds) as session:
                q = sql.insert(test_table1).values(id=5, title='test_title')
                await session.execute(q)
                q = sql.insert(test_table2).values(id=10, title='test_title2')
                await session.execute(q)
                raise Exception
        except:
            pass

        async with Session(engines, binds) as session:
            q = sql.select(test_table1.c).where(test_table1.c.id == 5)
            rows = await session.execute(q)
            self.assertEqual(rows.rowcount, 0)
            q = sql.select(test_table2.c).where(test_table1.c.id == 10)
            rows = await session.execute(q)
            self.assertEqual(rows.rowcount, 0)

    @asynctest
    async def test_commit(self, engines, binds):
        test_table1 = self.test_models['db1'].test_table1
        test_table2 = self.test_models['db2'].test_table2
        async with Session(engines, binds) as session:
            q = sql.insert(test_table1).values(id=5, title='test_title')
            await session.execute(q)
            q = sql.insert(test_table2).values(id=10, title='test_title2')
            await session.execute(q)
        async with Session(engines, binds) as session:
            q = sql.select(test_table1.c).where(test_table1.c.id == 5)
            rows = await session.execute(q)
            self.assertEqual(rows.rowcount, 1)
            q = sql.select(test_table2.c).where(test_table2.c.id == 10)
            rows = await session.execute(q)
            self.assertEqual(rows.rowcount, 1)

        try:
            async with Session(engines, binds) as session:
                q = sql.insert(test_table1).values(id=5, title='test_title')
                await session.execute(q)
                session.commit()
                q = sql.insert(test_table2).values(id=10, title='test_title2')
                await session.execute(q)
                session.commit()
                raise Exception
        except:
            pass

        async with Session(engines, binds) as session:
            q = sql.select(test_table1.c).where(test_table1.c.id == 5)
            rows = await session.execute(q)
            self.assertEqual(rows.rowcount, 1)
            q = sql.select(test_table2.c).where(test_table2.c.id == 10)
            rows = await session.execute(q)
            self.assertEqual(rows.rowcount, 1)


@skipIf(mysql_skip, 'Aiomysql not installed')
class MysqlSessionTestCase(_SessionTestCaseBase):
    db_url = cfg.MYSQL_URL


@skipIf(pg_skip, 'Postgress not installed')
class PgSessionTestCase(_SessionTestCaseBase):
    db_url = cfg.POSTGRESS_URL





from collections import OrderedDict

from .handlers import WebHandler
from .handlers import h_cases


class BaseView(object):

    name = None

    def __init__(self, env, data, **kwargs):
        self.env = env
        self.data = data
        self.namespace = env.namespace
        if not self.name:
            self.name = self.namespace.split('.')[-1]
        subreverse = env.root.build_subreverse(self.namespace,
                                               **data.as_dict())
        if not subreverse._ready:
            subreverse = subreverse(**data.as_dict())
        self.root = subreverse
        self.parent = getattr(env, 'view', None)
        self.__dict__.update(kwargs)

    @classmethod
    def cases(cls):
        return []

    @classmethod
    def handler(cls):
        return HView(cls) | h_cases(*cls.cases())

    @property
    def templates_folder(self):
        return self.name

    def template_name(self, template):
        folders = [self.templates_folder]
        parent = self.parent
        while parent:
            folders.insert(0, parent.templates_folder)
            parent = parent.parent
        template_path = '/'.join(folders)
        if template:
            template_path += '/{}'.format(template)
        return template_path

    def render_to_string(self, template, **kwargs):
        return self.env.render(
            self.template_name(template),
            **dict(kwargs, view=self)
        )

    def render_to_response(self, template, kwargs):
        return self.env.render.to_response(
            self.template_name(template),
            dict(kwargs, view=self),
        )

    def set_data(self, attr, value):
        setattr(self.env, '{}-{}'.format(self.name, attr), value)

    def get_data(self, attr, default):
        return getattr(self.env, '{}-{}'.format(self.name, attr), default)


class HView(WebHandler):
    def __init__(self, view_cls, **kwargs):
        self.view_cls = view_cls
        self.kwargs = kwargs

    def view(self, env, data):
        view = self.view_cls(env, data, **self.kwargs)
        views = getattr(env, 'views', {})
        view_by_cls = getattr(env, 'view_by_cls', {})
        #assert view.name not in views, view.name
        #assert view.__class__ not in view_by_cls
        views = OrderedDict(views)
        view_by_cls = OrderedDict(view_by_cls)
        views[view.name] = view
        view_by_cls[view.__class__] = view
        env.views = views
        env.view_by_cls = view_by_cls
        env.view = view
        return self.next_handler(env, data)
    __call__ = view


class HViewHandler(object):

    def __init__(self, method, view_cls=None):
        self.method = method
        self.view_cls = view_cls

    def __call__(self, env, data):
        assert self.view_cls
        view = env.view_by_cls[self.view_cls]
        return self.method(view, env, data)

    def __get__(self, instance, cls):
        if instance is None:
            return self.__class__(self.method, cls)
        else:
            return self.method.__get__(instance, cls)

viewhandler = HViewHandler


class HViewFilter(WebHandler):

    def __init__(self, method, view_cls=None):
        self.method = method
        self.view_cls = view_cls

    def __call__(self, env, data):
        assert self.view_cls
        next_handler = getattr(self, '_next_handler', None)
        view = env.view_by_cls[self.view_cls]
        return self.method(view, env, data, next_handler)

    def __get__(self, instance, cls):
        if instance is None:
            return self.__class__(self.method, cls)
        else:
            return self.method.__get__(instance, cls)

viewfilter = HViewFilter


class SafeProperty(object):

    def __init__(self, name, default=None, getter=None, setter=None):
        self.name = name
        self.default = default
        if getter:
            self.getter = getter
        if setter:
            self.setter = setter

    @staticmethod
    def getter(view, prop):
        return view.get_data(prop.name, prop.default)

    @staticmethod
    def setter(view, prop, value):
        view.set_data(prop.name, value)

    def __get__(self, view, view_cls):
        if view is None:
            return self
        else:
            return self.getter(view, self)

    def __set__(self, view, value):
        self.setter(view, self, value)




from __future__ import absolute_import
import logging

from sqlalchemy.dialects import registry
import ikcms.components.db.sqla

from .query import SphinxQuery


registry.register('sphinxql', 'ikcms.components.sphinx.dialect', 'Dialect')


class Component(ikcms.components.db.sqla.Component):
    name = 'sphinx'
    query_class = SphinxQuery
    xmlpipes = {}

    DEFAULT_SPHINX_URI = 'sphinxql://127.0.0.1:9306/?charset=utf8'
    DEFAULT_SPHINX_PARAMS = {
        'pool_size': 10,
        'max_overflow': 50,
        'pool_recycle': 3600,
    }

    @classmethod
    def create(cls, app):
        uri = getattr(app.cfg, 'SPHINX_URI', cls.DEFAULT_SPHINX_URI)
        params = getattr(app.cfg, 'SPHINX_PARAMS', cls.DEFAULT_SPHINX_PARAMS)
        engine = cls.create_engine('sphinx', uri, params)
        engine.logger = logging.getLogger('sqlalchemy.engine.[sphinx]')
        engines = {'sphinx': engine}
        models = {'sphinx': cls.get_models('sphinx')}
        return cls(app, engines, models)


component = Component.create_cls



# coding: utf8 
import os
import sys
import pkg_resources
import argparse
import shutil

import jinja2
import yaml
import jsonschema


def _resource_tree(package, resource):
    result = []
    for name in pkg_resources.resource_listdir(package, resource):
        path = os.path.join(resource, name)
        if pkg_resources.resource_isdir(package, path):
            result.append(name)
            subtree = _resource_tree(package, path)
            result += [os.path.join(name, x) for x in subtree]
        else:
            result.append(name)
    return result


def render(res, res_dir, target_dir='', kwargs={}):
    env = jinja2.Environment(
        loader=jinja2.PackageLoader(res, res_dir),
    )
    tree = _resource_tree(res, res_dir)
    for path in tree:
        target_path = os.path.join(target_dir, path)
        if pkg_resources.resource_isdir(res, os.path.join(res_dir, path)):
            os.mkdir(target_path)
        elif path.endswith('.j2'):
            target_path = target_path[:-3]
            if os.path.exists(target_path):
                print('Error: {} already exists'.format(target_path))
                return
            s = env.get_template(path).render(kwargs)
            with open(target_path, 'w') as f:
                f.write(s)
            print('{} created'.format(target_path))

class AppsCfg(dict):

    DEFAULT_FILEPATH = 'apps.yaml'

    schema = {
        "type": "object",
        "properties": {
            "apps" : {
                "type": "array",
                "items": {
                    "type": "string",
                },
            },
            "paths" : {
                "type": "array",
                "items": {
                    "type": "string",
                },
            },
        },
        "required": ["apps"],
    }

    def __new__(cls, **kwargs):
        self = dict.__new__(cls)
        self.update(kwargs)
        self.setdefault('apps', [])
        self.setdefault('paths', [])
        return self

    @classmethod
    def validate(cls, cfg):
        return jsonschema.validate(cfg, cls.schema)

    @classmethod
    def load(cls, filepath=None):
        try:
            with open(filepath or cls.DEFAULT_FILEPATH) as f:
                cfg = yaml.safe_load(f)
        except OSError:
            print("Can't open file {}".format(filepath))
            sys.exit()
        cls.validate(cfg)
        return cls(**cfg)


    def store(self, filepath=None):
        self.validate(self)
        try:
            with open(filepath or self.DEFAULT_FILEPATH, 'w') as f:
                yaml.dump(dict(self), f, default_flow_style=False)
        except OSError:
            print("Can't open file {}".format(filepath))
            sys.exit()


class Command(object):

    name = None
    description = 'some description'
    help = 'spme help'

    def __init__(self):
        assert self.name

    def args(self, parser):
        pass

    def __call__(self, **kwargs):
        raise NotImplementedError


class InitCommand(Command):

    name = 'init'
    description='init project'
    help='init project'


    def __call__(self, **kwargs):
        if os.path.exists(AppsCfg.DEFAULT_FILEPATH):
            print('Error: {} exists'.format(AppsCfg.DEFAULT_FILEPATH))
            sys.exit()
        AppsCfg().store()
        print('{} created'.format(AppsCfg.DEFAULT_FILEPATH))
        render('ikcms', 'ikinit/templates/init')


class AppCommand(Command):

    name = 'app'
    description='add simple app to project'
    help='add simple app to project'


    def args(self, parser):
        parser.add_argument('name')

    def __call__(self, **kwargs):
        name = kwargs['name']
        if os.path.exists(name):
            print('Error: {} already exists'.format(name))
            return
        os.mkdir(name)
        render('ikcms', 'ikinit/templates/app', name, dict(name=name))
        apps_cfg = AppsCfg.load()
        if name not in apps_cfg['apps']:
            apps_cfg['apps'].append(name)
        apps_cfg.store()


class CompositeCommand(Command):

    name = 'composite'
    description='add app with components to project'
    help='add app with components to project'


    def args(self, parser):
        parser.add_argument('name')
        parser.add_argument('--db', action='store_true')
        parser.add_argument('--jinja2', action='store_true')
        parser.add_argument('--memcache', action='store_true')
        parser.add_argument('--redis', action='store_true')
        parser.add_argument('--i18n', action='store_true')

    def __call__(self, **kwargs):
        name = kwargs['name']
        options = []
        if kwargs['db']:
            options.append('db')
        if kwargs['jinja2']:
            options.append('jinja2')
        if kwargs['memcache']:
            options.append('memcache')
        if kwargs['redis']:
            options.append('redis')
        if kwargs['i18n']:
            options.append('i18n')
        if os.path.exists(name):
            print('Error: {} already exists'.format(name))
            return
        os.mkdir(name)
        render(
            'ikcms',
            'ikinit/templates/composite',
            name,
            dict(options=options, name=name),
        )
        apps_cfg = AppsCfg.load()
        if name not in apps_cfg['apps']:
            apps_cfg['apps'].append(name)
        apps_cfg.store()


class AdminCommand(Command):

    name = 'admin'
    description='add admin app to project'
    help='add admin app to project'


    def args(self, parser):
        parser.add_argument('name', nargs='?', default='admin')

    def __call__(self, **kwargs):
        name = kwargs['name']
        if os.path.exists(name):
            print('Error: {} already exists'.format(name))
            return
        os.mkdir(name)

        render(
            'ikcms',
            'ikinit/templates/admin',
            name,
            dict(name=name),
        )

        src_static_js_dir = pkg_resources.resource_filename(
            'ikcms.apps.admin',
            'static/js',
        )
        dest_static_js_dir = os.path.join(name, 'static', 'js')
        shutil.copytree(src_static_js_dir, dest_static_js_dir)
        print('{} dir created'.format(dest_static_js_dir))

        src_static_css_dir = pkg_resources.resource_filename(
            'ikcms.apps.admin',
            'static/css',
        )
        dest_static_css_dir = os.path.join(name, 'static', 'css')
        shutil.copytree(src_static_css_dir, dest_static_css_dir)
        print('{} dir created'.format(dest_static_css_dir))


        src_cms_static_dir = pkg_resources.resource_filename(
            'iktomi.cms',
            'static',
        )
        dest_cms_static_dir = os.path.join(name, 'static/cms')
        shutil.copytree(src_cms_static_dir, dest_cms_static_dir)
        print('{} dir created'.format(dest_cms_static_dir))
        apps_cfg = AppsCfg.load()
        if name not in apps_cfg['apps']:
            apps_cfg['apps'].append(name)
        apps_cfg.store()



COMMANDS = [
    InitCommand(),
    AppCommand(),
    CompositeCommand(),
    AdminCommand(),
]


def cli():
    parser = argparse.ArgumentParser(
        description='Ikcms init script',
    )
    subparsers = parser.add_subparsers(
        title='commands',
        dest='command',
        help='commands',
        metavar='command',
    )
    parsers = {}
    for command in COMMANDS:
        parsers[command.name] = subparsers.add_parser(
            command.name,
            description=command.description,
            help=command.help,
        )
        command.args(parsers[command.name])

    parsers['help'] = subparsers.add_parser(
        'help',
        help='help for command',
        description='help for command',
    )
    parsers['help'].add_argument(
        'for_command',
        choices=[c.name for c in COMMANDS],
    )

    args = parser.parse_args()
    if args.command is None:
        parser.print_help()
    elif args.command == 'help':
        print(parsers[args.for_command].format_help())
    else:
        kwargs = vars(args).copy()
        del kwargs['command']
        for command in COMMANDS:
            if command.name == args.command:
                command(**kwargs)
                break





from ikcms.forms import fields
from ikcms.forms import convs


__all__ = (
    'name__required',
    'body',
    'handler',
    'handler__required',
    'request_id',
    'request_id__required',
    'error__required',
    'message__required',
    'body__error_required',
)

class name__required(fields.String):
    name = 'name'
    label = 'Название сообщения'
    raw_required = True
    required = True


class body(fields.RawDict):
    name = 'body'
    label = 'Тело сообщения'
    to_python_default = {}


class handler(fields.String):
    name = 'handler'
    label = 'Название хендлера'


class handler__required(handler):
    raw_required = True
    required = True


class request_id(fields.String):
    name = 'request_id'
    label = 'Идентификатор запроса'


class request_id__required(request_id):
    raw_required = True
    required = True


class error__required(fields.String):
    name = 'error'
    label = 'Идентификатор ошибки'
    raw_required = True
    required = True


class message__required(fields.String):
    name = 'message'
    label = 'Текстовое сообщение'
    raw_required = True
    required = True


class kwargs(fields.RawDict):
    name = 'kwargs'
    label = 'Ключевые аргументы'
    to_python_default = {}
    raw_required = False


class body__error_required(body):
    conv = convs.Dict
    fields = [
        error__required,
        message__required,
        kwargs,
    ]
    label = 'Teло сообщения об ошибке'
    raw_required = True
    required = True





import iktomi.web
from iktomi.web.reverse import Location

from .base import h_prefix
from .base import h_namespace


def h_subdomain(*subdomains, **kwargs):
    as_path = kwargs.pop('as_path', False)
    subdomain_handler = iktomi.web.subdomain(*subdomains, **kwargs)
    if as_path:
        primary = subdomain_handler.primary or subdomain_handler.domains[0]
        return h_prefix('/.{}'.format(primary), name=kwargs.get('name'))
    else:
        return subdomain_handler


class DomainLocation(Location):

    def __init__(self, *builders, **kwargs):
        super(DomainLocation, self).__init__(*builders, **kwargs)
        self.default_domain = kwargs.get('default_domain')

    def build_subdomians(self, reverse):
        subdomain = super(DomainLocation, self).build_subdomians(reverse)
        if reverse._bound_env:
            domain = reverse._bound_env.domain
        else:
            domain = self.default_domain
        return ".".join(filter(None, [subdomain, domain]))


class HDomain(iktomi.web.WebHandler):

    def __init__(self, domains=None, primary=None, name=None, default=None):
        super(HDomain, self).__init__()
        self.domains = domains
        self.primary = primary
        self.default = default or primary or domains and domains[0]
        if name is not None:
            self._next_handler = h_namespace(name)

    def domain(self, env, data):
        request_domain = env._route_state.subdomain
        if self.domains is not None:
            matched_domain = self._match_domain_by_specified(
                request_domain,
                self.domains,
            )
        else:
            matched_domain = self._match_domain_by_subdomains(
                request_domain,
                self.subdomains(env),
            )
        if matched_domain:
            env.domain = self.primary or matched_domain
            env._route_state = env._route_state.add_subdomain(
                self.primary or matched_domain,
                matched_domain,
            )
            return self.next_handler(env, data)
        else:
            return None
    __call__ = domain

    def subdomains(self, env):
        if not hasattr(self, '_subdomains'):
            self._subdomains = set()
            for key, (location, scope) in self._locations().items():
                self._subdomains.update(
                    self._collect_subdomains(location, scope),
                )
        return self._subdomains

    def _match_domain_by_specified(self, request_domain, specified_domains):
        for domain in specified_domains:
            slen = len(domain)
            delimiter = request_domain[-slen - 1:-slen]
            if request_domain.endswith(domain) and delimiter in ('', '.'):
                return domain

    def _match_domain_by_subdomains(self, request_domain, subdomains):
        for subd in subdomains:
            if request_domain.startswith(subd + '.'):
                return request_domain[len(subd) + 1:]
        else:
            return request_domain

    def _collect_subdomains(self, location, scope):
        collected = set()
        if location and location.subdomains:
            subdomain = location.subdomains[-1]
            if hasattr(subdomain, 'subdomains'):
                for s in subdomain.subdomains:
                    if s:
                        collected.add(s)
                    elif s is None and scope:
                        for loc, sc in scope.values():
                            collected |= self._collect_subdomains(loc, sc)
            else:
                collected.add(subdomain)
        elif scope:
            for loc, sc in scope.values():
                collected |= self._collect_subdomains(loc, sc)
        return collected

    def _locations(self):
        locations = super(HDomain, self)._locations()
        new_locations = {}
        for key, (location, scope) in locations.items():
            location = DomainLocation(
                *location.builders,
                subdomains=location.subdomains,
                default_domain=self.default
            )
            new_locations[key] = (location, scope)
        return new_locations


def h_domain(
    domains=None,
    primary=None,
    name=None,
    default=None,
    as_path=False,
):
    domain_handler = HDomain(domains, primary, name, default)
    if as_path:
        if domain_handler.domains:
            return h_prefix('/.{}'.format(domain_handler.default), name=name)
        else:
            return h_prefix('')
    else:
        return domain_handler





from unittest import TestCase, skipIf
from unittest.mock import MagicMock
try:
    from ikcms.components.cache.memcache import component
    skip_test = False
except ImportError:
    skip_test = True

from tests.cfg import cfg


@skipIf(skip_test, 'Memcache not installed')
class MemcacheTestCase(TestCase):

    def test_cache(self):
        app = self._create_app()

        cache = component().create(app)
        cache.delete('test_key')

        value = cache.get('test_key')
        self.assertIsNone(value)

        cache.set('test_key', 'test_value')
        value = cache.get('test_key')
        self.assertEqual(value, 'test_value')

        cache.delete('test_key')
        value = cache.get('test_key')
        self.assertIsNone(value)

        app = self._create_app()

        cache_with_key = component(prefix='test_prefix-').create(app)
        cache.delete('test_prefix-test_key')
        cache_with_key.set('test_key', 'test_value')

        value = cache.get('test_prefix-test_key')
        self.assertEqual(value, 'test_value')
        value = cache_with_key.get('test_key')
        self.assertEqual(value, 'test_value')
        cache_with_key.delete('test_key')
        value = cache.get('test_prefix-test_key')
        self.assertIsNone(value)

    def _create_app(self):
        app = MagicMock()
        del app.cache
        app.cfg = cfg
        return app



from ikcms.forms import fields


class login(fields.String):
    name = 'login'
    label = 'Логин'
    raw_required = True
    required = True


class password(fields.String):
    name = 'password'
    label = 'Пароль'
    raw_required = True
    required = True



import six
from iktomi.db.sqla.types import Html

from lxml import html
from jinja2 import Markup


class ExpandableMarkup(object):
    '''
        Wrapper for markup which should be preprocessed on front-end.
    '''

    def __init__(self, markup):
        if isinstance(markup, ExpandableMarkup):
            self.markup = markup.markup
        else:
            self.markup = Markup(markup)

    def __len__(self):
        return len(self.markup)

    def __unicode__(self):
        raise RuntimeError('ExpandableMarkup is not converted and ' +
                           'can not be displayed')
    __html__ = __str__ = __unicode__

    def __eq__(self, other):
        if isinstance(other, ExpandableMarkup):
            return self.markup == other.markup
        if isinstance(other, six.string_types):
            return self.markup == other
        return False


class ExpandableHtml(Html):
    '''
    Column type for markup which should be preprocessed on front-end.
    Made in purpose to ensure that source markup is not occasianally
    outputted on front.
    If it is outputted directly, RuntimeError is raised.
    Use it with corresponding form converter.
    '''

    markup_class = ExpandableMarkup

    def process_bind_param(self, value, dialect):
        if value is not None:
            return six.u(value.markup)




import aiomcache
from . import base


class Component(base.Component):

    DEFAULT_MEMCACHE_HOST = 'localhost'
    DEFAULT_MEMCACHE_PORT = 11211
    prefix = b''

    def __init__(self, app, memcache):
        super().__init__(app)
        self.memcache = memcache

    @classmethod
    async def create(cls, app):
        host = getattr(app.cfg, 'MEMCACHE_HOST', cls.DEFAULT_MEMCACHE_HOST)
        port = getattr(app.cfg, 'MEMCACHE_PORT', cls.DEFAULT_MEMCACHE_PORT)
        memcache = aiomcache.Client(host, port)
        return cls(app, memcache)

    async def get(self, key):
        return await self.memcache.get(self._key(key))

    async def set(self, key, value, expire=0):
        return await self.memcache.set(self._key(key), value, exptime=expire)

    async def delete(self, key):
        return await self.memcache.delete(self._key(key))

    def _key(self, key):
        return self.prefix + key


component = Component.create_cls



import ikcms.components.render.jinja2
import ikcms.components.db.sqla
import ikcms.apps.composite


class App(ikcms.apps.composite.App):

    components = [
        ikcms.components.render.jinja2.component(paths=['{SITE_DIR}/templates']),
        ikcms.components.db.sqla.component(),
    ]

    def get_handler(self):
        raise NotImplementedError



from iktomi.web import WebHandler
from .. import exceptions


class h_guard(WebHandler):
    '''
        params:    None - do not check anything
                         dict - {key: type}

        methods:  if request.method not in methods, throws MethodNotAllowed
    '''

    def __init__(self, methods=('GET',), params=()):
        if params == '*':
            self.params = params
        else:
            self.params = dict(params)

        methods = list(methods)
        if 'GET' in methods and 'HEAD' not in methods:
            methods.append('HEAD')
        self.methods = methods

    def check_method(self, request):
        return  request.method in self.methods

    def check_params(self, request):
        if self.params == '*':
            return True

        checked_args = set()
        # Unwrap CacheableRequest
        if hasattr(request, 'unwrap'):
            request = request.unwrap()
        for key, value in request.GET.items():
            if key.startswith('utm_') or key.startswith('hc_'):
                continue
            if key in checked_args or key not in self.params:
                return False
            checked_args.add(key)
            tp = self.params[key]
            if type(tp) in (list, tuple):
                if value not in tp:
                    return False
            elif tp is not None and tp != "":
                try:
                    tp(value)
                except ValueError: # XXX write validation
                    return False
        return True

    def guard(self, env, data):
        if not self.check_method(env.request):
            raise exceptions.HTTPMethodNotAllowed
        if not self.check_params(env.request):
            raise exceptions.HTTPNotFound
        return self.next_handler(env, data)
    __call__ = guard





#!/usr/bin/env python
# -*- coding: utf-8 -*-
from ikcms.cli.base import Cli

class GeneratorCli(Cli):
    name = 'generator'

    def command_run(self, *names):
        app = self.create_app()
        app.db.generate(names)



import re

import sqlalchemy.sql.expression
from sqlalchemy.sql.visitors import replacement_traverse
from sqlalchemy.sql.base import _generative
from sqlalchemy.util.langhelpers import public_factory


_escape_re = re.compile(r'([=\(\)|\-!@~\"&/\\\^\$\=])')


def escape(value):
    return _escape_re.sub(r'\\\1', value)


class OptionsClause(sqlalchemy.sql.expression.ClauseElement):
    __visit_name__ = 'options'

    def __init__(self, options):
        self.options = options

    def extend(self, options):
        self.options.update(options)

    def _clone(self):
        return OptionsClause(self.options.copy())

    def __iter__(self):
        return self.options.iteritems()


class MatchClause(sqlalchemy.sql.expression.ClauseElement):
    __visit_name__ = 'match'

    def __init__(self, values, fields):
        self.values = list(values)
        self.fields = fields

    def extend(self, args, kwargs):
        self.values.extend(list(args))
        self.fields.update(kwargs)

    def _clone(self):
        return MatchClause(self.values[:], self.fields.copy())


class Select(sqlalchemy.sql.expression.Select):
    _options = None

    @_generative
    def match(self, *args, **kwargs):
        clause_already_exists = [False]

        def replace(node):
            if isinstance(node, MatchClause):
                clause_already_exists[0] = True
                node.extend(args, kwargs)
            return node

        self._whereclause = replacement_traverse(self._whereclause, {}, replace)
        if not clause_already_exists[0]:
            self.append_whereclause(MatchClause(args, kwargs))

    @_generative
    def options(self, **kwargs):
        if self._options is None:
            self._options = OptionsClause(kwargs)
        else:
            self._options.extend(kwargs)


select = public_factory(Select, '.expression.extract')



import time
from hashlib import md5


class Client:

    def __init__(self, app, server, client_id):
        self.app = app
        self._server = server
        self._client_id = client_id
        self.session_id = self.get_session_id()

    def get_session_id(self):
        salt = self.app.cfg.WS_AUTH_SECRET
        raw_session_id = "{}.{}.{}".format(time.time(), self._client_id, salt)
        return md5(raw_session_id.encode('utf8')).hexdigest()

    def remote_address(self):
        return self._server.get_remote_address(self._client_id)

    async def close(self):
        pass




import sys
from setuptools import setup
from setuptools import find_packages

try:
    from babel.messages import frontend as babel
    BABEL = True
except:
    BABEL = False

PY3 = sys.version_info >= (3,0)

cmdclass = {}
if BABEL:
    cmdclass.update({
        'extract_messages': babel.extract_messages,
        'init_catalog': babel.init_catalog,
        'update_catalog': babel.update_catalog,
    })


setup(
    name='ikcms',
    version='0.6',
    packages=find_packages(),
    package_dir={'ikcms': 'ikcms'},
    test_suite='nose.collector',
    tests_require=['nose'],
    install_requires=[
        'six',
        'pytz',
        'babel',
        'webob', #necessary for iktomi
        'sqlalchemy', #necessary for iktomi
        'iktomi',
        'PyYAML',
        'jinja2', #necessary for ikinit
        'jsonschema',
    ],
    include_package_data=True,
    extras_require={
        'mysql': ['PyMySQL'],
        'aiomysql': ['aiomysql'],
        'aiopg': ['aiopg'],
        'memcache': [
            PY3 and 'pymemcache' or 'pymemcache',
        ],
        'aiomcache': ['aiomcache'],
        'redis': ['redis'],
        'aioredis': ['aioredis'],
    },
    package_data={'ikcms': [
        'ikinit/templates/*/*.j2',
        'locale/*/*.po',
        'apps/admin/static/css/Manifest',
        'apps/admin/static/css/*.css',
        'apps/admin/static/js/Manifest',
        'apps/admin/static/js/*.js',
        'apps/admin/static/js/lib/*.js',
    ]},
    cmdclass=cmdclass,
    message_extractors={
        'ikcms': [('**.py', 'python', None)],
    },
    entry_points = {
        'console_scripts': [
            'ikinit = ikcms.ikinit:cli',
        ],
    },
)



import importlib

from sqlalchemy.engine.url import make_url

from iktomi.utils import cached_property

import ikcms.ws_components.base
from ikcms import orm


class Component(ikcms.ws_components.base.Component):

    name = 'db'
    session_cls = orm.Session

    def __init__(self, app, engines):
        super().__init__(app)
        self.engines = engines

    @classmethod
    async def create(cls, app):
        databases = getattr(app.cfg, 'DATABASES', {})
        database_params = getattr(app.cfg, 'DATABASE_PARAMS', {})
        engines = {}
        for db_id, url in databases.items():
            engines[db_id] = await cls.create_engine(db_id, url, database_params)
        return cls(app, engines)

    @classmethod
    async def create_engine(cls, db_id, url, engine_params=None):
        engine_params = engine_params or {}
        sa_url = make_url(url)
        module_name = '.{}'.format(sa_url.drivername)
        module = importlib.import_module(module_name, __package__)
        #assert sa_url.drivername in cls.drivers, \
        #    'Unknown db driver {}'.format(sa_url.drivername)
        return await module.create_engine(sa_url, **engine_params)

    @cached_property
    def mappers(self):
        from models.mappers import registry
        return registry

    @cached_property
    def binds(self):
        binds = {}
        for db_id, metadata  in self.mappers.metadata.items():
            for table in metadata.sorted_tables:
                binds[table] = self.engines[db_id]
        return binds

    async def __call__(self):
        return self.session_cls(self.engines, self.binds)

    async def close(self):
        for engine in self.engines.values():
            engine.terminate()
            await engine.wait_closed()


component = Component.create_cls




import ikcms.apps.base.env


class Environment(ikcms.apps.base.env.Environment):

    def __init__(self, *args, **kwargs):
        super(Environment, self).__init__(*args, **kwargs)
        for component in self.app.components:
            component.on_init_env(self)

    def close(self):
        for component in self.app.components:
            component.on_close_env(self)



from unittest import TestCase
from unittest.mock import MagicMock
from ikcms.forms.validators import *
from ikcms.forms import exceptions


class RequiredTestCase(TestCase):
    def test_is_required(self):
        field = MagicMock()
        field.required = True
        field.required_error = 'Field is required'
        validator = Required(field)
        for value in ['', [], {}, None]:
            with self.assertRaises(exceptions.ValidationError) as ctx:
                validator(value)
            exc = ctx.exception
            self.assertEqual(exc.kwargs['error'], field.required_error)

    def test_is_not_required(self):
        field = MagicMock()
        field.required = False
        field.required_error = 'Field is required'
        validator = Required(field)
        for value in ['', [], {}, None]:
            validator(value)


class RegexTestCase(TestCase):
    def test_regex(self):
        field = MagicMock()
        field.regex = r'\d{4}-\d{2}-\d{2}'
        field.regex_error = 'Field does not match regex'
        validator = Regex(field)
        with self.assertRaises(exceptions.ValidationError) as ctx:
            validator('invalid')
        self.assertEqual(ctx.exception.kwargs['error'], field.regex_error)
        validator('2016-06-15')


class LenTestCase(TestCase):
    def test_len(self):
        field = MagicMock()
        field.min_len = 1
        field.max_len = 10
        field.min_len_error = 'Minimum length error'
        field.max_len_error = 'Maximum length error'
        validator = Len(field)
        validator(list(range(1)))
        validator(list(range(5)))
        validator(list(range(10)))
        with self.assertRaises(exceptions.ValidationError) as ctx:
            validator(list(range(0)))
        self.assertEqual(ctx.exception.kwargs['error'], field.min_len_error)
        with self.assertRaises(exceptions.ValidationError) as ctx:
            validator(list(range(11)))
        self.assertEqual(ctx.exception.kwargs['error'], field.max_len_error)


class RangeTestCase(TestCase):
    def test_range(self):
        field = MagicMock()
        field.min_value = 1
        field.max_value = 10
        field.min_value_error = 'Minimum value error'
        field.max_value_error = 'Maximum value error'
        validator = Range(field)
        validator(1)
        validator(5)
        validator(10)
        with self.assertRaises(exceptions.ValidationError) as ctx:
            validator(0)
        self.assertEqual(ctx.exception.kwargs['error'], field.min_value_error)
        with self.assertRaises(exceptions.ValidationError) as ctx:
            validator(11)
        self.assertEqual(ctx.exception.kwargs['error'], field.max_value_error)




class Widget:

    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)

    @property
    def name(self):
        return self.__class__.__name__

    def to_dict(self, field):
        return {
            'widget': self.name,
            'name': field.name,
            'label': field.label,
            'fields': [f.widget.to_dict(f) for f in field.named_fields.values()],
        }




from __future__ import absolute_import
import json
from functools import partial
from os import path

from iktomi.cms.item_lock.redis import RedisItemLock as ItemLock
from iktomi.utils import quote_js
from iktomi.utils.storage import (
    storage_method,
    storage_cached_property,
)

import ikcms.apps.composite.env
from ikcms.utils import cached_property
from ikcms.web import Response

class Environment(ikcms.apps.composite.env.Environment):

    def render_to_response(self, *args, **kwargs):
        return self.render.to_response(*args, **kwargs)

    @storage_method
    def json(self, data):
        if getattr(self, '_flash', None):
            assert '_flashmessages' not in data
            data = dict(data, _flashmessages=self._flash)
        return Response(
            json.dumps(data),
            content_type="application/json",
            charset='utf8',
        )

    @cached_property
    def cfg(self):
        return self.app.cfg

    @cached_property
    def streams(self):
        return self.app.streams

    @storage_cached_property
    def url_for(storage):
        return storage.root.build_url

    def url_for_static(self, part):
        while part.startswith('/'):
            part = part[1:]
        return path.join(self.cfg.STATIC_URL, part)

    def render_to_string(self, template_name, context):
        return self.render.render(template_name, **context)

    @cached_property
    def render_to_response(self):
        return self.render.to_response

    @cached_property
    def template(self):
        return self.render

    @cached_property
    def context(self):
        return self.Context(self)

    #@storage_cached_property
    #def user(storage):
    #    return storage.user

    @cached_property
    def models(self):
        class Models(object):
            admin = self.app.db.get_models('admin')
            front = self.app.db.get_models('front')
        return Models()

    @storage_cached_property
    def redis(storage):
        return storage.app.cache.client

    @cached_property
    def auth_model(self):
        return self.models.admin.AdminUser

    @storage_cached_property
    def item_lock(storage):
        return ItemLock(storage)

    @storage_cached_property
    def lang(storage):
        return storage.app.i18n.Lang(storage.app.i18n, 'ru')

    @storage_cached_property
    def gettext(self):
        return self.lang.gettext

    @storage_cached_property
    def ngettext(self):
        return self.lang.ngettext

    @storage_method
    def get_template_vars(self):
        d = dict(
            context=self.app.get_context(self),
            user=getattr(self, 'user', None),
            quote_js=quote_js,
            url_for=self.url_for,
            url_for_static=self.url_for_static,
            packed_js_tag=partial(self.app.packer.js_tag, self),
            packed_css_tag=partial(self.app.packer.css_tag, self),
        )
        import_settings = getattr(self.cfg, 'TEMPLATE_IMPORT_SETTINGS', [])
        for key in import_settings:
            d[key] = getattr(self.cfg, key)
        return d

    @storage_cached_property
    def _lang(storage):
        lang = getattr(storage, 'lang', 'ru')
        return storage.app.i18n.langs[lang]

    @storage_cached_property
    def gettext(storage):
        return storage._lang.gettext

    @storage_cached_property
    def ngettext(storage):
        return storage._lang.ngettext

    # XXX: Copypasted from iktomi-cms Environment
    # override this method with configured HelpLoader.get_help to be able
    # to use help messages in admin
    def get_help(self, *args, **kwargs):
        return ''




import ikcms.ws_components.base
from . import exceptions
from . import streams


__all__ = (
    'component',
    'exceptions',
    'streams',
)


class Component(ikcms.ws_components.base.Component):

    name = 'streams'
    streams = {}

    roles = {
        'streams.read': 'Потоки: чтение',
        'streams.edit': 'Потоки: редактирование',
    }

    def __init__(self, app):
        super().__init__(app)
        registry = {}
        for cls in self.streams:
            cls.create(self, registry)
        self.streams = registry

    def get_cfg(self, env):
        return {
            'streams': [s.get_cfg(env) for s in self.streams.values()],
        }

    async def h_action(self, env, message):
        stream_name = message.get('stream')
        stream = self.streams.get(stream_name)
        if stream:
            return await stream.h_action(env, message)
        else:
            raise exceptions.ClientError(exceptions.StreamNotFound(stream_name))


component = Component.create_cls



# -*- codung: utf8 -*-
from jinja2 import nodes, Markup
from jinja2.ext import Extension

from ikcms.utils import cached_property

class Tag(Extension):

    tags = None
    template = None

    def __init__(self, environment):
        assert self.tags
        assert self.template
        super(Tag, self).__init__(environment)

    def parse(self, parser):
        output = self.call_method('render', kwargs=self._parse_kwargs(parser))
        return nodes.Output([output])

    def _parse_kwargs(self, parser):
        lineno = next(parser.stream).lineno
        # XXX Is env required?
        #env = nodes.Name()
        #env.name = 'env'
        #env.ctx = 'load'
        #kwargs = [nodes.Keyword('env', env)]
        kwargs = []
        while parser.stream.current.type!='block_end':
            if len(kwargs)>1:
                parser.stream.expect('comma')
            key = parser.stream.expect('name').value
            parser.stream.expect('assign')
            value = parser.parse_expression()
            kwargs.append(nodes.Keyword(key, value))
        return kwargs

    def render(self, *args, **kwargs):
        l = self.func(**kwargs)
        if l is None:
            return ''
        else:
            template = self.template.format(**l)
            html = self.environment.get_template(template).render(**l)
            return Markup(html)

    @classmethod
    def create(cls, template):
        def wrapper(func):
            return type(
                func.func_name,
                (cls,),
                dict(
                    tags=set([func.func_name]),
                    func=staticmethod(func),
                    template=template,
                ),
            )
        return wrapper



class TagWithBody(Tag):

    def parse(self, parser):
        kwargs = self._parse_kwargs(parser)
        body = self._parse_body(parser)
        call = self.call_method('render', kwargs=kwargs)
        return nodes.CallBlock(call, [], [], body)

    def _parse_body(self, parser):
        return parser.parse_statements(
            ['name:end%s' % list(self.tags)[0]],
            drop_needle=True,
        )


class CustomTags(list):

    def tag(self, template):
        def wrapper(func):
            tag = Tag.create(template)(func)
            self.append(tag)
            return tag
        return wrapper

    def tag_with_body(self, template):
        def wrapper(func):
            tag = TagWithBody.create(template)(func)
            self.append(tag)
            return tag
        return wrapper




from datetime import date
from unittest import TestCase
from unittest.mock import MagicMock
from ikcms.forms.convs import *
from ikcms.forms import exceptions
from ikcms.forms import fields


class _BaseConvTestCase(TestCase):

    conv_class = None
    test_values = []
    raw_type_error_values = []
    python_type_error_values = []
    validation_error_values = []

    def setUp(self):
        field = MagicMock(not_null=False, name='test_name')
        self.conv = self.conv_class(field)

    def test_to_python(self):
        for python_value, raw_value in self.test_values:
            self.assertEqual(
                python_value,
                self.conv.to_python(raw_value),
                self._to_python_message(python_value, raw_value),
            )
        for value in self.raw_type_error_values:
            self.assertRaises(
                exceptions.RawValueTypeError,
                self.conv.to_python,
                value,
            )

        for value in self.validation_error_values:
            self.assertRaises(
                exceptions.ValidationError,
                self.conv.to_python,
                value,
            )

    def test_not_none(self):
        self.assertRaises(
                exceptions.RawValueNoneNotAllowedError,
            self.conv.to_python,
            None,
        )
        self.assertRaises(
            exceptions.PythonValueNoneNotAllowedError,
            self.conv.from_python,
            None,
        )
        self.conv.field.not_none = False
        self.assertEqual(
            None,
            self.conv.to_python(None),
            self._to_python_message(None, None),
        )
        self.assertEqual(
            None,
            self.conv.from_python(None),
            self._from_python_message(None, None),
        )

    def test_from_python(self):
        for python_value, raw_value in self.test_values:
            self.assertEqual(
                raw_value,
                self.conv.from_python(python_value),
                self._from_python_message(python_value, raw_value),
            )
        for value in self.python_type_error_values:
            self.assertRaises(
                exceptions.PythonValueTypeError,
                self.conv.from_python,
                value,
            )

    def _to_python_message(self, raw_value, value):
        return "{}().to_python({})!={}".format(
            self.conv_class.__name__,
            raw_value,
            value,
        )

    def _from_python_message(self, value, raw_value):
        return "{}().from_python({})!={}".format(
            self.conv_class.__name__,
            value,
            raw_value,
        )


class BoolConvTestCase(_BaseConvTestCase):

    conv_class = Bool
    test_values = [
        (True, True),
        (False, False),
    ]
    raw_type_error_values = [0, 1, -5, 7.6, '', 'aaa', [], {}, [1, 3]]
    python_type_error_values = raw_type_error_values


class IntConvTestCase(_BaseConvTestCase):
    conv_class = Int
    test_values = [
        (-1, -1),
        (0, 0),
        (10, 10),
    ]
    raw_type_error_values = ['23', 'sss', 4.5, [1, 2], {1: 1}]
    python_type_error_values = raw_type_error_values


class IntStrConvTestCase(_BaseConvTestCase):
    conv_class = IntStr
    test_values = [
        (-1, '-1'),
        (0, '0'),
        (10, '10'),
    ]
    raw_type_error_values = [23, 4.5, [1, 2], {1: 1}]
    python_type_error_values = ['23', 'sss', 4.5, [1, 2], {1: 1}]
    validation_error_values = ["2.2", "dddd", "fr45g", "-7-"]


class StrConvTestCase(_BaseConvTestCase):
    conv_class = Str
    test_values = [
        ("", ""),
        ("11", "11"),
        ("aaa123", "aaa123"),
    ]
    raw_type_error_values = [True, False, 5, 5.6, [], {}, [1, 3]]
    python_type_error_values = raw_type_error_values


class ListConvTestCase(_BaseConvTestCase):

    conv_class = List
    test_values = [
        ([], []),
        ([1], [1]),
        ([1, 2], [1, 2]),
    ]
    raw_type_error_values = [
        True, False, 5, 5.6, "test", {}, {"test":[]},
    ]
    python_type_error_values = raw_type_error_values

    def setUp(self):
        field = MagicMock(
            not_null=False,
            name='test_name',
            fields=[fields.Int()],
        )
        self.conv = self.conv_class(field)


class DictConvTestCase(_BaseConvTestCase):

    conv_class = Dict
    test_values = [
        (
            {'str_field': 'str', 'int_field': 1},
            {'str_field': 'str', 'int_field': 1},
        ),
    ]
    raw_type_error_values = []
    python_type_error_values = raw_type_error_values


    def setUp(self):
        str_field = fields.Field()
        str_field.name = 'str_field'
        str_field.conv = Str(str_field)
        str_field.raw_required = False
        str_field.to_python_default = 'str_default'
        int_field = fields.Field()
        int_field.name = 'int_field'
        int_field.conv = Int(int_field)
        int_field.raw_required = False
        int_field.to_python_default = 'int_default'
        int_field.to_python = int_field.to_python
        field = fields.Field()
        field.fields = [str_field, int_field]
        self.conv = Dict(field)


class DateConvTestCase(_BaseConvTestCase):
    conv_class = Date
    test_values = [
        (date(2016, 6, 15), '2016-06-15'),
    ]

    raw_type_error_values = [True, False, 5, 5.6, [], {}, [1, 3]]
    python_type_error_values = raw_type_error_values

    def setUp(self):
        field = MagicMock(
            not_null=False,
            name='test_name',
            format='%Y-%m-%d',
        )
        self.conv = self.conv_class(field)




import ikcms.components.base


class Component(ikcms.components.base.Component):

    name = 'cache'

    @property
    def WatchError(self):
        raise NotImplementedError

    def get(self, key):
        raise NotImplementedError

    def mget(self, *keys):
        raise NotImplementedError

    def set(self, key, value, expires=0):
        raise NotImplementedError

    def mset(self, mapping):
        raise NotImplementedError

    def add(self, key, value, expires=0):
        raise NotImplementedError

    def delete(self, *keys):
        raise NotImplementedError

    def pipe(self):
        raise NotImplementedError






import asyncio
import logging

import websockets

from .base import ServerBase


logger = logging.getLogger(__name__)


class WS_Server(ServerBase):

    sockets = {}

    def __init__(self, host, port, app):
        self.host = host
        self.port = port
        self.app = app

    def serve_forever(self):
        start_server = websockets.serve(
            self._new_client,
            self.host,
            self.port,
        )
        asyncio.get_event_loop().run_until_complete(start_server)
        asyncio.get_event_loop().run_forever()

    async def _new_client(self, websocket, path):
        logger.debug('Connected client %s', websocket.remote_address)

        client_id = self.client_id(websocket)
        self.sockets[client_id] = websocket
        logger.debug('Connected client %s', websocket.remote_address)
        try:
            await self.app(self, client_id)
        except websockets.exceptions.ConnectionClosed:
            logger.debug('Disconnected client %s', websocket.remote_address)
        except:
            await self.disconnect(client_id, 500, 'Internal server error')
            raise

    def client_id(self, websocket):
        return websocket.__repr__()

    def get_remote_address(self, client_id):
        return self.sockets[client_id].remote_address

    async def send(self, client_id, data):
        return await self.sockets[client_id].send(data)

    async def recv(self, client_id):
        return await self.sockets[client_id].recv()

    async def disconnect(self, client_id, code=1000, reason=''):
        address = self.get_remote_address(client_id)
        await self.sockets[client_id].close(code, reason)
        del self.sockets[client_id]
        logger.debug('Disconnected client %s', address)

    async def ping(self, client_id):
        return self.sockets[client_id].ping()




import logging

import ikcms.web
from iktomi.utils.storage import VersionedStorage
from iktomi.web.app import is_host_valid
from iktomi.web import Reverse

import ikcms.cli.app


logger = logging.getLogger(__name__)


class App(object):

    Request = ikcms.web.Request
    Response = ikcms.web.Response

    HTTPException = ikcms.web.exceptions.HTTPException
    HTTPNotFound = ikcms.web.exceptions.HTTPNotFound
    HTTPSeeOther = ikcms.web.exceptions.HTTPSeeOther
    HTTPForbidden = ikcms.web.exceptions.HTTPForbidden
    HTTPInternalServerError = ikcms.web.exceptions.HTTPInternalServerError
    HTTPMethodNotAllowed = ikcms.web.exceptions.HTTPMethodNotAllowed

    commands = {
        'app': ikcms.cli.app.AppCli,
    }

    def __init__(self, cfg):
        self.cfg = cfg
        self.cfg.config_uid()
        self.cfg.config_logging()
        self.env_class = self.get_env_class()
        self.handler = self.get_handler()
        self.root = self.get_root()

    def get_request(self, environ):
        return self.Request(environ, charset='utf-8')

    def get_env_class(self):
        from .env import Environment
        return Environment

    def get_env(self, request):
        return VersionedStorage(self.env_class, request=request, app=self)

    def get_data(self):
        return VersionedStorage()

    def get_handler(self):
        raise NotImplementedError

    def get_root(self):
        return Reverse.from_handler(self.handler)

    def handle_error(self, env):
        '''
        Unhandled exception handler.
        You can put any logging, error warning, etc here.'''
        logger.exception('Exception for %s %s :',
                         env.request.method, env.request.url)

    def handle(self, env, data):
        '''
        Calls application and handles following cases:
            * catches `webob.HTTPException` errors.
            * catches unhandled exceptions, calls `handle_error` method
              and returns 500.
            * returns 404 if the app has returned None`.
        '''
        try:
            try:
                response = self.handler(env, data)
                if response is None:
                    logger.debug('Application returned None '
                                 'instead of Response object')
                    response = self.HTTPNotFound()
            finally:
                env.close()
        except self.HTTPException as e:
            response = e
        except Exception as e:
            self.handle_error(env)
            response = self.HTTPInternalServerError()
        return response

    def __call__(self, environ, start_response):
        '''
        WSGI interface method.
        Creates webob and iktomi wrappers and calls `handle` method.
        '''
        # validating Host header to prevent problems with url parsing
        if not is_host_valid(environ['HTTP_HOST']):
            logger.warning(
                'Unusual header "Host: %s", return HTTPNotFound',
                environ['HTTP_HOST'])
            return self.HTTPNotFound()(environ, start_response)

        request = self.get_request(environ)
        try:
            env = self.get_env(request)
            data = self.get_data()
            response = self.handle(env, data)
            result = response(environ, start_response)
        except Exception:
            self.handle_error(env)
            result = self.HTTPInternalServerError()(environ, start_response)
        return result




from webob.exc import HTTPException
from webob.exc import HTTPNotFound
from webob.exc import HTTPSeeOther
from webob.exc import HTTPForbidden
from webob.exc import HTTPInternalServerError
from webob.exc import HTTPMethodNotAllowed




from sqlalchemy.schema import MetaData
from sqlalchemy import Column
from sqlalchemy import Integer
from sqlalchemy import String
from sqlalchemy import Date
from sqlalchemy import sql
from sqlalchemy.ext.declarative import declarative_base




class ModelsBase:
    async def reset(self, conn):
        for table in self.metadata.sorted_tables[::-1]:
            try:
                await conn.execute(sql.ddl.DropTable(table))
            except Exception:
                pass
        for table in self.metadata.sorted_tables:
            await conn.execute(sql.ddl.CreateTable(table))


def create_models1():
    class Models1(ModelsBase):
        db_id = 'db1'
        metadata = MetaData()
        Base = declarative_base(metadata=metadata)

        class Test(Base):
            __tablename__ = 'Test'
            id = Column(Integer, primary_key=True)
            title = Column(String(60), nullable=True)
            title2 = Column(String(60), nullable=True)
            date = Column(Date, nullable=True)

        test_table1 = Test.__table__
    return Models1()

def create_models2():
    class Models2(ModelsBase):
        db_id = 'db2'
        metadata = MetaData()
        Base = declarative_base(metadata=metadata)

        class Test2(Base):
            __tablename__ = 'Test2'
            id = Column(Integer, primary_key=True)
            title = Column(String(60), nullable=True)
            title2 = Column(String(60), nullable=True)

        test_table2 = Test2.__table__
    return Models2()


def create_metadata(models):
    return {model.db_id: model.metadata for model in models}




import ikcms.forms

from .. import exceptions


class Form(ikcms.forms.Form):

    def to_python(self, raw_value, keys=None):
        try:
            return super().to_python(raw_value, keys=keys)
        except exceptions.RawValueError as exc:
            raise exceptions.MessageError({
                exc.kwargs['field_name']: exc.kwargs['error'],
            })




import json
from ikcms.orm import mappers


class Draft(mappers.Base):

    name = 'Draft'
    db_id = 'admin'

    async def get_new_item_draft(self, session, stream_id, user_id):
        return await self.query().filter_by(
            user_id=user_id,
            stream_id=stream_id,
            item_id=None,
        ).select_first_item(session)

    async def store_new_item_draft(self, session, stream_id, user_id, values):
        data = self.from_python(values)
        draft = await get_new_item_draft(session, stream_id, user_id)
        if draft:
            draft['data'] = data
            await self.query().update_item(session, draft['id'], draft)
        else:
            draft = {
                'stream_id': stream_id,
                'user_id': user_id,
                'data': data,
            }
            await self.query().insert_item(session, draft)

    async def get_edit_item_draft(self, session, stream_id, item_id):
        return await self.query().filter_by(
            stream_id=stream_id,
            item_id=item_id,
        ).select_first_item(session)

    async def store_edit_item_draft(self, session, stream_id, item_id, values):
        data = self.from_python(values)
        draft = await get_edit_item_draft(session, stream_id, item_id)
        if draft:
            draft['data'] = data
            await self.query().update_item(session, draft['id'], draft)
        else:
            draft = {
                'stream_id': stream_id,
                'item_id': item_id,
                'data': data,
            }
            await self.query().insert_item(session, draft)

    async def delete_draft(self, item_id):
        with self.app.db() as sesion:
            await self.query().delete_item(session, item_id)

    def to_python(self, raw_data):
        return json.loads(raw_data)

    def from_python(self, python_data):
        return json.dump(raw_data)


